<?php
/**
 * @copyright © 2008, 2009 Intrahealth International, Inc.
 * This File is part of I2CE
 *
 * I2CE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
*  I2CE_CustomReport
* @package I2CE
* @subpackage Core
* @author Carl Leitner <litlfred@ibiblio.org>
* @version 2.1
* @access public
*/


class I2CE_CustomReport extends I2CE_Fuzzy{
    
    /**
     * The magic data node holding the configuration information for this report
     * $var protected I2CE_MagicDataNode $config
     */
    protected $config;



    /**
     * $var protected string $table the name of the cached  table of the report
     */
    protected $table;

    /**
     * $var protected string $tmp_table the temporary name of the cached  table of the report
     */
    protected $tmp_table;

    /**
     * @var protected MDB2 $db The instance of the database to perform queries on.
     */
    protected $db;
    

    /**
     * @var protected array $populate_queries   an array of queries used to populate the cached reports table(s)
     */
    protected $populate_queries;

    /**
     * @var protected array $init_queries   an array of queries used to inialize the cached reports table(s)
     */
    protected $init_queries;

    /**
     * The constructor
     * @param string $report  The report name
     */
    public function __construct($report ) {
        $this->report = $report;
        if (!$this->reportExists($report)) {
            $msg = "The requested report $report does not exist";
            I2CE::raiseError($msg);
            throw new Exception($msg); 
        }
        $this->config = I2CE::getConfig()->traverse("/modules/CustomReports/reports/$report");
        $this->relationship = '';
        $this->config->setIfIsSet($this->relationship,'relationship');
        if (!$this->relationship) {
            $msg = "The requested report $report does not have a relationship";
            I2CE::raiseError($msg);
            throw new Exception($msg); 
        }
        $this->db = MDB2::singleton();
        $this->table = self::getCachedTableName($report,true);
        $this->tmp_table =  self::getCachedTableName($report,true,'tmp_custom_report');
        $this->init_queries = array();
        $this->populate_queries = array();
        $qry = "SELECT * FROM `information_schema`.`COLUMNS` WHERE `TABLE_SCHEMA` = '" . I2CE_CachedForm::getCacheDatabase() .
            "' AND `TABLE_NAME` =  ? AND `COLUMN_NAME` = ?";
        $this->get_field_def = $this->db->prepare($qry,array('text','text'),MDB2_PREPARE_RESULT);
        $msg = "Unable to prepare field definition query";
        if (I2CE::pearError($this->get_field_def,$msg)) {
            throw new Exception($msg); 
        }
        //this might throw an exception.  don't catch it here.
        $this->rel = new I2CE_FormRelationship($this->relationship, '/modules/CustomReports/relationships', array('I2CE_CachedForm','getCachedTableName'));
    }


    public function getFormRelationship() {
        return $this->rel;
    }

    /**
     * @var protected MDB2_PreparedStatement $get_field_def
     */
    protected $get_field_def;
    

    protected static $report_table_cols = array();

    /**
     * Get the actual columns in the cached report table
     */
    public static function getColumnsInReportTable($report) {
        if (array_key_exists($report, self::$report_table_cols) && is_array(self::$report_table_cols[$report])) {
            return self::$report_table_cols[$report];
        }
        $cols_in_report = array();
        $cols_qry = 
            "SELECT column_name FROM information_schema.columns WHERE "
            . " CONCAT('`',TABLE_SCHEMA,'`.`',TABLE_NAME , '`') = '" . self::getCachedTableName($report,true) . "'";
        $db = MDB2::singleton();
        $res = $db->query($cols_qry);
        if (I2CE::pearError($res,"Could not get columns in report " .  self::getCachedTableName($report,true))) {
            return array();
        } 
        while ($row = $res->fetchRow()) {
            if (  PEAR::isError( $row ) ) {
                break;
            }
            $cols_in_report[] = $row->column_name;
        }
        self::$report_table_cols[$report] = $cols_in_report;
        return self::$report_table_cols[$report];
    }


    /**
     * Gets the forms required by the report
     * @param string $report
     * @returns array of string
     */
    public function getFormsRequiredByReport() {
        if (!$this->rel instanceof I2CE_FormRelationship) {
            return array();
        }
        return $this->rel->getRequiredForms();
    }

    /**
     * Below is the static code to actually handle checking/setting the status of a cached report
     * 
     */


    public static function getCachedTableName($report,$withDB = true, $table_prefix = '') {
        $db_name = '';
        $DBConfig = I2CE::getConfig()->modules->CustomReports->database_options;
        if ($withDB) {
            $DBConfig->setIfIsSet($db_name,'database');
            if (strlen($db_name) > 0) {
                if ($db_name[strlen($db_name) -1] !== '.') {
                    $db_name = '`' . $db_name . '`.';
                } else {
                    $db_name = '`' . substr($db_name,0,-1) . '`.';
                }            
            } else {
                $db_name =         '`'  . MDB2::singleton()->database_name . '`.';
            }
        }        
        if (!$table_prefix) {
            $table_prefix = '';
            $DBConfig->setIfIsSet($table_prefix,'table_prefix');
        }
        if (strlen($table_prefix) > 0) {
            if ($table_prefix[strlen($table_prefix)-1] !== '_') {
                $table_prefix .= '_';
            }
        }
        return $db_name . '`' .   $table_prefix . $report  . '`';
    }


    /**
     * Checks to see if a report exists
     * @param string $report
     * @returns boolean
     */
    public static function  reportExists($report) {
        if (!I2CE_MagicDataNode::checkKey($report)) {
            return false;
        }
        return I2CE::getConfig()->__isset("/modules/CustomReports/reports/$report");
    }


    /**
     * Checks to see if a report has failed in its generation
     * @param string $report
     * @returns mixed.  True if it has failed, false if it has not failed, null if the report does not exist.
     */
    public static function hasFailed($report) {
        if (!self::reportExists($report)) {
            return null;
        }
        $status =  self::getStatus($report);
        if ($status === 'failed') {
            return true;
        }
        if (($status !== 'in_progress') || ($status !== 'stale')) {
            return false;
        }
        $timeConfig = I2CE::getConfig()->traverse('/modules/CustomReports/times',true);        
        $timeConfig->volatile(true);
        $fail_time  = null;
        $timeConfig->setIfIsSet($fail_time,'fail');
        if (!is_integer($fail_time)) {
            $fail_time = 600;
        }
        $fail_time = (int)( ((int)$fail_time) * 60);        
        $begin_generation = self::getLastGenerationTime($report);
        if (  (time() - ((int) $begin_generation) ) >   $fail_time)  {
            self::setStatus($report,'failed');
            return true;
        }
        return false;        
    }

    protected static $hooman = array( //really should be moved to a template file so we can translate.  i'm too lazy
        'does_not_exist'=>'Does not exist',
        'not_generated'=>'Not generated',
        'stale'=>'Stale',
        'failed'=>'Failed',
        'in_progress'=>'In Progress',
        'generated'=>'Generated'

        );

    /**
     * Get the status of the requested report
     * @param string $report
     * @param boolean hooman_readable defaults to false.
     * @returns string 'does_not_exist','not_generated','generated','failed','in_progress', 'stale' if it is not hooman readable
     **/
    public static  function getStatus($report,$hooman_readable=false) {
        if (!self::reportExists($report)) {
            if ($hooman_readable) {
                return $hooman['does_not_exist'];
            } else {
                return 'does_not_exist';
            }
        }
        $status = '';
        I2CE::getConfig()->modules->CustomReports->status->volatile(true);
        I2CE::getConfig()->setIfIsSet($status, "/modules/CustomReports/status/$report");
        if ($status) {
            if ($hooman_readable) {
                return self::$hooman[$status];
            } else {
                return $status;
            }
        } else {
            self::setStatus($report,'not_generated');
            if ($hooman_readable) {
                return self::$hooman['not_generated'];
            } else {
                return 'not_generated';
            }
        }
    }

    /**
     * Set the status us a report.
     * @param string $report
     * @param string $status
     */
    public static function setStatus($report,$status) {
        if (!self::reportExists($report)) {
            return;
        }
        I2CE::getConfig()->__set("/modules/CustomReports/status/$report", $status);
    }

    /**
     * Check to see if a report is stale.
     * @param string $report
     * @returns boolean true/false 
     */
    public  static function isStale($report) {
        if (!self::reportExists($report)) {
            return true;
        }
        $status =  self::getStatus($report);
        if ($status === 'stale' || $status == 'failed' || $status == 'not_generated') {
            return true;
        }
        if ($status === 'in_progress') {
            return false;
        }

        $generation_time = self::getLastGenerationTime($report);
        if ($generation_time === null || $generation_time <= 0) {
            return true;
        }
        if ($generation_time > time()) { //make sure there is no time sync. problem
            return true;
        }
        //status is not in progress
        $timeConfig = I2CE::getConfig()->traverse('/modules/CustomReports/times',true);        
        $timeConfig->volatile(true);
        $stale_time  = 10;
        if ($timeConfig->setIfIsSet($stale_time,'stale')) {        
            if (is_integer($stale_time) || (is_string($stale_time) && ctype_digit($stale_time))) {
                if ($stale_time <= 0) {
                    return true;
                }
            } else {
                $stale_time = 10;
            }
        }
        //lookup per-reprot stale time and override if  necc.
        $t_stale_time = null;
        if ($timeConfig->setIfIsSet($t_stale_time,"stale_by_report/$report")) {
            if (is_integer($t_stale_time) ||  (is_string($stale_time) && ctype_digit($stale_time))) {
                if ( $t_stale_time > 0 ) {
                    $stale_time = $t_stale_time;
                } else {
                    return true;
                }
            }
        }
        $stale_time = $stale_time * 60; //convert to minutes
        return  (($generation_time + $stale_time) < time()); 
    }


    /**
     *@param string $report The shortname for the report
     *@return mixed.  False on failure, int the time the last report generation  on sucess
     */
    public static function getLastGenerationTime($report) {
        if (!self::reportExists($report)) {
            return false;
        }
        $timeConfig = I2CE::getConfig()->traverse('/modules/CustomReports/times',true);        
        $timeConfig->volatile(true);
        $generation = 0;
        $timeConfig->setIfIsSet($generation,"generation/$report");        
        if (  (!(is_integer($generation) || ctype_digit($generation))) || ((int)$generation) < 1) {
            self::setStatus($report,'not_generated');
            return false;
        }
        return $generation;

    }


    /**
     * Below is the code to actually handle generation of reports
     * 
     */

    /**
     *  Generate the cached report
     * @param boolean $force.  Defaults to false.  If set to true, it will force the regeneration of the report if it is in_progress
     * @param boolean $cache_forms.  Defaults to True.  If set to true, it will cache the forms required by this report
     */
    public function generateCache($force=false,$cache_forms = true) {
        if (!self::reportExists($this->report)) {
            return false;
        }
        if ((!$force) && $this->getStatus($this->report) == 'in_progress') {
            I2CE::raiseError("Cached report in progress for {$this->report} and generation is not forced");
            return false;
        }
        if (!$force && !self::isStale($this->report)) {
            I2CE::raiseError("Report {$this->report} is not stale");
            //report is not stale
            return true;
        }
        if ($cache_forms) {
            $forms = $this->getFormsRequiredByReport();
            I2CE::raiseError("Attempting to cache the required forms for the report {$this->report}:\n\t" . implode(',', $forms));
            $failures = array();
            foreach ($forms as $form) {
                try {
                    $cachedForm = new I2CE_CachedForm($form);
                }
                catch(Exception $e) {
                    if (array_key_exists('HTTP_HOST',$_SERVER)) { //we don't need to check here, b/c it won't error out.  we are doing it to keep the log file clean
                        $this->userMessage ( "Unable to setup cached form $form");
                    }
                    $failures[] = $form;
                    continue;
                }
                if (!$cachedForm->generateCachedTable(!$force)) {
                    $failures[] = $form;
                    continue;
                }
            }
            if (count($failures) > 0) {
                I2CE::raiseError("Warning data may be out of date for report {$this->report} -- could not cache forms:\n\t" . implode(',' ,$failures));
            } else {
                I2CE::raiseError("Successfully cached all forms");
            }
        }
        $this->setStatus($this->report,'in_progress');
        $timeConfig = I2CE::getConfig()->traverse('/modules/CustomReports/times',true);        
        $timeConfig->volatile(true);
        $timeConfig->generation->{$this->report} = time();
        if (!$this->_generateCache()) {
            $this->setStatus($this->report,'failed');
            I2CE::raiseError("Report generation failed for {$this->report}");
            return false;
        }
        I2CE::raiseError("Report generation succeeded for {$this->report}");
        $this->setStatus($this->report,'generated');
        $timeConfig->generation->{$this->report} = time();
        return true;
    }
    

    



    protected function _generateCache() {
        if (!$this->setupQueries()) {
            I2CE::raiseError("Could not process form tree for report {$this->report}");
            return false;
        } 
        foreach ($this->init_queries as $qry) {
            I2CE::raiseError("Executing:\n$qry");
            if (I2CE::pearError($this->db->exec($qry), "Unable to initialize  cached report")) {
                return false;
            }        
        }
        foreach ($this->populate_queries as $qry) {
            I2CE::raiseError("Doing $qry");
            $res = $this->db->exec($qry);
            if (I2CE::pearError($res, "Unable to populate cached report")) {
                return false;
            }      
            I2CE::raiseError("Inserted $res rows");
        }
        $this->processReportingFunctions(); //does not report any failures.   trys to recover.
        // we move obver the tmp_table totable.
        if (I2CE::pearError($this->db->exec("DROP TABLE IF EXISTS {$this->table}"), "Could not drop old table {$this->table}")) {
            return false;
        }
        if (I2CE::pearError($this->db->exec("RENAME TABLE {$this->tmp_table} TO {$this->table}"),"Could not rename temp table")) {
            return false;
        }
        return true;
    }

    protected $reportedFunctions;


    protected function processReportingFunctions() {
        $reportedFunctions = $this->getReportedFunctions();
        foreach ($reportedFunctions as $function=>$data) {
            $qry = "INSERT INTO {$this->tmp_table} (`primary_form+id`,`+$function`) SELECT `primary_form+id` , {$data['qry']} AS `+$function`"
                    . " FROM {$this->tmp_table} ON DUPLICATE KEY UPDATE `+$function`=values(`+$function`)";
            I2CE::raiseError("Doing -- " . $qry);
            $res = $this->db->exec( $qry );
            if (I2CE::pearError( $res, "Warning: unable to populate `+$function` by {$data['qry']}" )) {
                continue;
            }
            I2CE::raiseError( "Inserted $res rows" );
        }
    }

    protected function getReportedFunctions() {
        if (is_array($this->reportedFunctions)) {
            return $this->reportedFunctions;
        }
        $this->reportedFunctions = array();
        if (!$this->config->is_parent('reporting_functions')) {
            return  $this->reportedFunctions;
        }
        foreach ( $this->config->reporting_functions as $function=>$functionConfig) {
            if (!$functionConfig instanceof I2CE_MagicDataNode) {
                I2CE::raiseError("Skipping  $function b/c not magic data node");
                continue;
            }
            if (!$functionConfig->is_scalar('enabled') || !$functionConfig->enabled) {
                I2CE::raiseError("Skipping  $function b/c not enabled");
                continue;
            }
            if ( ($details = $this->rel->getFunctionDetails($function)) === false) {
                continue;
            }
            $this->reportedFunctions[$function] = $details;
        }        
        return $this->reportedFunctions;
    }


    protected function setupQueries() {
        $formConfigs = $this->rel->getFormConfig();
        $parentConfigs = $this->rel->getParentFormNames();
        $primaryForm = $this->rel->getPrimaryForm();
        $primaryFormReference = $this->rel->getReferencedForm($primaryForm);
        if (!$primaryFormReference) {
            I2CE::raiseError("Could not get a reference for the primary form $primaryForm");
            return false;
        }
        $primaryFormName = $this->rel->getPrimaryFormName();
        $create_fields = array();
        $create_fields ["primary_form+id"] = "`primary_form+id` varchar(255)";
        $create_fields ["primary_form+parent"] = "`primary_form+parent` varchar(255)";
        $keys = array('PRIMARY KEY (`primary_form+id`)');
        //get the required fields for reported functions
        $requiredFields = array();
        $reportingFunctions = $this->getReportedFunctions();
        foreach ($reportingFunctions as $functions=>$data) {
            foreach ($data['required_fields'] as $reportformfield) {
                list($reportform,$field) = explode('+',$reportformfield);
                if (!$reportform || !$field || $field == 'id' ) {
                    continue;
                }
                if (!array_key_exists($reportform,$requiredFields)) {
                    $requiredFields[$reportform] = array();
                }
                $requiredFields[$reportform][$field] = true;
            }
        }
        $formNames = $this->rel->getFormNames();
        $factory = I2CE_FormFactory::instance();
        foreach ($formNames as $formName) {
            $form = $this->rel->getForm($formName);
            $formReference = $this->rel->getReferencedForm($form);
            $formObj = $factory->createForm($form);
            if (!$formObj instanceof I2CE_Form) {
                I2CE::raiseError("$formName in report references $form which could no be instantiated");
                return false;
            }
            $parentFormName = $this->rel->getParentFormNames($formName);
            $parentForm = $this->rel->getForm($parentFormName);
            //$parentFormReference = $this->rel->getReferencedForm($parentForm);
            if (!$form || !$parentForm || !$parentFormName || !$formReference ) {
//            if (!$form || !$parentForm || !$parentFormName || !$formReference || !$parentFormReference ) {
                I2CE::raiseError("Invalid form data at $formName: $form/$parentForm/$parentFormName/$formReference/$parentFormReference");
                return false;
            }
            $reporting_fields =    $this->rel->getLimitingFields($formName); //get the fields used in limits.            
            foreach ($this->rel->getChildFormNames($formName) as $childFormName) {
                $reporting_fields = array_merge($reporting_fields,$this->rel->getJoiningFields($childFormName));
            }
            $limit_fields = $reporting_fields;
            if (array_key_exists($formName,$requiredFields)) {
                $reporting_fields = array_merge($reporting_fields,$requiredFields[$formName]);
            }
            if ('primary_form' == $formName) {
                $reportConfig = $this->config->reporting_forms->primary_form;
                if (array_key_exists('primary_form',$requiredFields)) {
                    $reporting_fields = array_merge($reporting_fields,$requiredFields['primary_form']);
                } 
            }else {
                $reportConfig = $this->config->reporting_forms->$formName;
            }
            $reportingFieldsBase = $reportConfig->fields;
            foreach ($reportingFieldsBase  as $field=>$fConfig) {
                if (is_scalar($fConfig)) {
                    continue;
                }
                if (!isset($fConfig->enabled)) {
                    continue;
                }
                //now check that this field is actual in the form
                $fieldObj = $formObj->getField($field);
                if ($fConfig->enabled) {                    
                    $reporting_fields[$field]=true;
                    continue;
                }
                if (!$fConfig->is_parent('limits')) {
                    continue;
                }
                foreach ($fConfig->limits as $limit=>$data) {
                    if ((!$data instanceof I2CE_MagicDataNode)  || !$data->is_scalar('data') || !$data->enabled) {
                        continue;
                    }
                    $reporting_fields[$field] = true;  
                }
            }                        
            //now we go through all forms looking for link appends that refer to this form.
            foreach ($this->config->reporting_forms as $linkFormName=>$linkConfig) {
                foreach ($linkConfig->fields as $linkField=>$fConfig) {
                    if (!(isset($fConfig->link_append) && $fConfig->is_scalar('link_append') &&  ($fConfig->link_append))) {
                        continue;
                    }
                    if (strpos($fConfig->link_append,'+') !== false) {
                        list($link_form,$link_field) = explode('+', $fConfig->link_append);
                    } else {
                        $link_form = $linkFormName;
                        $link_field = $fConfig->link_append;
                    }
                    if (! (($link_form == $formName) || ($link_form == 'primary_form' && $formName == $primaryFormName) )) {
                        continue;
                    }
                    if ($link_field && $link_field == 'id') {
                        continue;
                    }
                    $reporting_fields[$link_field] = true;   
                }                
            }
            $as = array();
            $updates = array();
            if ($formName != 'primary_form') {
                $as[ "`primary_form+id`" ] = "`primary_form+id`";
                //$as[ "`{$parentFormName}+id`"] = "`{$parentFormName}+id`";
                $as[ "`{$formName}+id`"] = "`$formName`.id  as `{$formName}+id`";
                $create_fields["$formName+id"] = "`{$formName}+id` varchar(255)";
                $keys[] = "INDEX (`{$formName}+id`)";
                //the next line is just to be safe. i think we only need this in the case that $parentFormName == $primaryFormName
                //$updates[] = "`{$parentFormName}+id` = values(`{$parentFormName}+id`)"; 
                $updates[] = "`{$formName}+id` = values(`{$formName}+id`)"; 
            } else {
                $as[ "`primary_form+id`"] = "`$parentForm`.id as `primary_form+id`";
                $as[ "`primary_form+parent`"] = "`$parentForm`.parent as `primary_form+parent`";
            }
            

//             if ('primary_form' == $formName && count($reporting_fields) == 0 ) {
//                 continue; //no fields were added from the primary field so we may skip population from the primary form
//             }

            $fields = array_keys($reporting_fields);
            foreach ($fields as $field) {
                if (!is_string($field) || strlen($field) == 0) {
                    I2CE::raiseError("Invalid field data for $formName");
                    continue;
                }
                $fieldObj = $formObj->getField($field);
                if (!$fieldObj instanceof I2CE_FormField) {
                    I2CE::raiseError("$formName / $form in report references the field $field in $form which could no be instantiated:" .get_class($formObj), E_USER_ERROR);
                    continue;
                }
                $name = "$formName+$field";
                if (array_key_exists($name,$as)) {
                    I2CE::raiseError("Reported field $field as $name is duplicated for {$this->report} at $formName");
                    continue;
                }
                $create_field = $this->getCreateField($form,$field, $name);
                if ($create_field === false) {                    
                    I2CE::raiseError("Unable to find field data for $form.$field in {$this->report}");
                    continue;
                } 
                $create_fields[$name] = $create_field['field'];
                if (in_array($field, $limit_fields)) { //this field is used as a limit so lets try and put an index
                    $type = $create_field['type'];
                    if (preg_match('/^varchar\(([0-9]*)\)$/',$type,$matches)) {
                        $amt = min($matches[1],255);
                        $keys[] = "INDEX `$name` (`$name` ($amt))";
                    }else if (preg_match('/^char\(([0-9]*)\)$/',$type,$matches)) {
                        $amt = min($matches[1],255);
                        $keys[] = "INDEX `$name` (`$name` ($amt))";
                    } else if ($type == 'text') {
                        $keys[] = "INDEX `$name` (`$name` (255))";
                    } else if ($type == 'longtext') {
                        $keys[] = "INDEX `$name` (`$name` (255))";
                    } else if (preg_match('/blob/',$type)) {
                        //do nothing
                    } else if (preg_match('/^set/',$type)) {
                        //do nothing
                    } else if (preg_match('/^enum/',$type)) {
                        //do nothing
                    } else {
                        $keys[] = "INDEX `$name` (`$name`)";
                    }
                }
                if ($formName !== 'primary_form') {
                    $as['`'.$name.'`'] = "`$formName`.`$field`  AS `$name` ";
                } else {
                    $as['`'.$name.'`'] = "`$form`.`$field`  AS `$name` ";
                }
                if ($formName != $primaryFormName) {
                    $updates[] = "`$name`=values(`$name`)";
                }
            }


            if ($formName == 'primary_form') {
                $join = '';
                $from = " FROM " .  $primaryFormReference . " AS `$primaryForm` ";
            } else {
                $join = $this->rel->getJoin($formName, $parentForm);
                if ($join === false) {
                    I2CE::raiseError("Unable to join $formName");
                    continue;
                }
                //. $join . " JOIN {$this->tmp_table}  ON `$parentForm`.id = `{$parentFormName}+id` ";
                //$join . " JOIN {$this->tmp_table}  ON `$parentForm`.id = `{$parentFormName}+id` ";
                //$from = " FROM $formReference AS `$formName` ";
                $from = " FROM  {$this->tmp_table} ";
            } 
            
            $where = $this->rel->generateWhereClause($formName);
            if (!is_string($where)) {
                I2CE::raiseError("Could not get where clause for $formName=>$form for {$this->report}");
                $where = '';
            }  
            $where =trim($where);
            if (strlen($where) > 0) {
                $where = ' WHERE (' . $where . ')';
            }
            if (count($updates) > 0) { 
                $update= " ON DUPLICATE KEY UPDATE " . implode(',', $updates);
            }
            $this->populate_queries[] =
                "INSERT INTO {$this->tmp_table} ( " 
                . implode(',', array_keys($as)) . ")" 
                . " SELECT " . implode(',',$as) 
                . $from
                //. " FROM " . $parentFormReference . " AS `{$parentForm}` "
                .  $join .  $where . $update;
            if ($this->rel->dropEmpty($formName)) {
                $this->populate_queries[] = "DELETE FROM {$this->tmp_table} WHERE `{$formName}+id` IS NULL";
            }
        }
        //now add in any reported functions.
        foreach ($reportingFunctions as $function=>$data) {
            $create_fields["+$function"] = "`+$function` {$data['type']}";
        }
        $this->init_queries[] = "DROP TABLE IF EXISTS {$this->tmp_table}";        
        $create_fields = array_merge($create_fields, $keys);
        $this->init_queries[]  = "CREATE TABLE {$this->tmp_table} ( "  .  implode(',', $create_fields) . ")  ENGINE=InnoDB DEFAULT CHARSET=utf8 COLLATE=utf8_bin";        
        return true;
    }




    protected function getCreateField($form,$field,$name) {
        $result = $this->get_field_def->execute(array(I2CE_CachedForm::getCachedTableName($form,false),$field));
        if (I2CE::pearError($result,"Unable to get field defintion for " . I2CE_CachedForm::getCachedTableName($form,false) . ".$field")) {
            return false;
        }
        if ($result->numRows() != 1) {
            I2CE::raiseError("Unexpected number of rows " . $result->numRows() . " when determining field defintion for $form.$field");
            return false;
        }
        $result = $result->fetchRow();
        $field = "`$name` " . $result->column_type . ' ' ;
        if ($result->character_set_name){
            $field .= ' CHARACTER SET ' . $result->character_set_name . ' ';
        } 
        if (isset($result->collation_name)) {
            $field .= ' COLLATE ' . $result->collation_name;
        }
        return array('type'=> $result->column_type, 'field'=>$field);
    }

  

}
# Local Variables:
# mode: php
# c-default-style: "bsd"
# indent-tabs-mode: nil
# c-basic-offset: 4
# End:
