<?php
  /**
 * @copyright © 2008, 2009 Intrahealth International, Inc.
 * This File is part of I2CE
 *
 * I2CE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
   */
  /**
   *  I2CE_CustomReport_Display
   * @package I2CE
   * @subpackage Core
   * @author Carl Leitner <litlfred@ibiblio.org>
   * @version 2.1
   * @access public
   */


abstract class I2CE_CustomReport_Display extends I2CE_Fuzzy{

    /**
     * an array indexed by   forms 
     * @var protected array $formObjs 
     */
    protected $formObjs;
    

    /**an array which tells us of if a report field  is mapped.  Keys are report fields of the form "$reportform+$field(+$aggregate)". values are booleans
     * @var protected array mappedFields
     */
    protected $mappedFields;

    /**
     * An array where keys are of the  "$reportform(+field(+$aggregate))" and the values are the corresponding "$form+$field" (that is the real form for the later)
     * @var protected array $maps
     */
    protected $formMaps;
    /**
     * An array where keys are of the  "$reportform(+field(+$aggregate))" and the values are the corresponding "$form+$field" (that is the real form for the later)
     * @var protected array $fieldMaps
     */
    protected $fieldMaps;
    
    /**an array of mapped values indexed by "$form+$field" (Note: $form is _not_ the report form, but the true form shortname)
     * @var protected array mappedValues
     */
    protected $mappedValues;
    
    /**
     * The shortname for the display.  E.g. 'Default'
     * @var protected string $display
     */
    protected $display;

    /**
     * @var protected I2CE_CustomReport
     */
    protected $reportObj;
    
    /**
     * The constuctor
     * @param I2CE_Page $page
     * @param string $view
     * @throws Excecption on error
     */
    public function __construct($page,$view) {
        $this->page = $page;
        $this->template= $this->page->getTemplate();
        $this->fieldMaps = array();
        $this->formMaps = array();
        $this->mappedFields = array();
        $this->mappedValues = array();
        $this->formObjs = array();
        $this->display = 'Default';
        if (preg_match('/^I2CE_CustomReport_Display_([a-zA-Z0-9_]+)$/',get_class($this),$matches)) {
            $this->display = $matches[1];
        }
        $config = I2CE::getConfig()->modules->CustomReports;
        $this->displayConfig = $config->displays->{$this->display};
        if (!isset($config->reportViews->$view)) {
            throw new Exception("You specified an invalid report view ($view)");
        }
        $this->view = $view;        
        $this->config = $config->reportViews->$view;
        if (!I2CE_CustomReport::reportExists($this->config->report)) {
            throw new Exception("View refers to nonexistent report {$this->config->report}");
        }
        $status = I2CE_CustomReport::getStatus( $this->config->report );
        if ( !$status || $status == 'not_generated' || $status == 'failed' ) {
            throw new Exception("Report for {$this->config->report} has not been generated");
        }
        $this->reportObj = new I2CE_CustomReport($this->config->report); //may throw an error. don't catch it

        $this->reportConfig = $config->reports->{$this->config->report};
        if (!$this->reportConfig instanceof I2CE_MagicDataNode) {
            throw new Exception("Report  {$this->config->report} is invalid");
        }
        if (!isset($this->reportConfig->relationship) || !$this->reportConfig->relationship) {
            throw new Exception("View refers to nonexistent relationship {$this->reportConfig->relationship}");
        }
        $this->relationshipConfig = $config->relationships->{$this->reportConfig->relationship};
        if (!$this->relationshipConfig instanceof I2CE_MagicDataNode) {
            throw new Exception("Report relationship {$this->reportConfig->relationship} is invalid");
        }
        $this->defaultOptions = $this->getDefaultOptions($this->page->request(),array() );        
        if (!is_array($this->defaultOptions)) {  
            throw new Exception("Cannot get display options");
        }
        if (!$this->canView()) {
            throw new Exception("Cannot look at view {$this->view}");
        }        

    }
    
    /**
     * The options for this display
     */
    protected $defaultOptions;
    /**
     * the view we are displaying
     * @var protected string $view
     */
    protected $view;
    /**
     * The page we are displaying on
     * @var protected I2CE_Page $page
     */
    protected $page;
    /**
     * @var protected I2CE_TemplateMeister $template
     */
    protected $template;

    
    /**
     * The magic data node which holds the information about the display that is being used
     * @var protected I2CE_MagicDataNode $displayConfig
     */
    protected $displayConfig;
    
    /**
     * The magic data node which holds the information about this view
     * @var protected I2CE_MagicDataNode $config
     */
    protected $config;
    /**
     * The magic data node which holds the information about the report this view referes to
     * @var protected I2CE_MagicDataNode $reportConfig
     */
    protected $reportConfig;

    /**
     * The magic data node which holds the information about the relationship this view refers to
     * @var protected I2CE_MagicDataNode $relationshipConfig
     */
    protected $relationshipConfig;

    /**
     * Abstract method that each display is resposbile for implementing.  Checks to see
     * if it can display the given view.
     * @returns boolean
     */
    abstract protected function canView();



    /**
     * array  with keys the 'reportformfields' and values the data associated to it.  
     * these fields are those we assume have a numeric representation
     * @var protected array $numeric
     */
    protected $numeric;


        
    /**
     * Find the numeric fields.  Returns it (and stores in in the variable $this->numeric)
     * @returns @array.  Keys are report for fields and values are the data associate defined in getDisplayedFields()
     */
    protected function findNumericFields( $disabled = false ) {
        if ($disabled == false && is_array($this->numeric)) {
            return $this->numeric;
        }
        $numeric = array();
        // get disabled fields. get none of the aggregate information
        $reportformfields = $this->getReportViewDisplayedFields( $disabled, array( '' ) );
        //now we need to see if we have at least one numeric data column and one text data column
        foreach ($reportformfields as $reportformfield=>$data) {
            list( $form, $field ) = array_pad(explode( '+', $reportformfield ),2,'');
            if ( strpos( $field, '+' ) !== false ) {
                continue; //badness we somehow have aggregate type info
            }
            if (!$form) { //this is a function field
                $is_numeric = false;
                $this->relationshipConfig->setIfIsSet($is_numeric,"reporting_functions/$field/is_numeric");
                if ($is_numeric) {
                    $numeric[$reportformfield] = $data; //$reportformfield == $field in this case
                }
                continue;
            }
            list($formObj,$fieldObj) = $this->getFormFieldObjects($reportformfield);
            if (!$formObj instanceof I2CE_Form) {
                continue;
            }
            if ($formObj->isNumeric($field)) {
                $numeric[$reportformfield] = $data;
                continue;
            }
        }
        if ( $disabled === false ) {
            $this->numeric = $numeric;
        }
        return $numeric;
    }


    /**
     * @param array $limits an nested array defining limiting values for particular fields. Defaults to the empty array. 
     * @returns mixed false on failure on succes an array. at index 'results' and  MDB2 buffered result object  at index 'num_results' the
     * number of results that would be found without the limit
     */
    protected function getResults() {
        if (array_key_exists('limit_paginated',$this->defaultOptions) && $this->defaultOptions['limit_paginated']) {
            if (!array_key_exists('limit_page', $this->defaultOptions) || !(is_integer($this->defaultOptions['limit_page']) || ctype_digit($this->defaultOptions['limit_page'])) || ((int)$this->defaultOptions['limit_page']) < 1) {
                $this->defaultOptions['limit_page'] = 1;
            }
            if (!array_key_exists('limit_per_page',$this->defaultOptions) || ! (is_integer($this->defaultOptions['limit_per_page']) || ctype_digit($this->defaultOptions['limit_per_page'] ))|| ((int)$this->defaultOptions['limit_per_page']) < 1) {
                //we don't have a valid 'limit_per_page'
                $this->defaultOptions['limit_per_page'] = 100; //default to 100 
            }
            $this->defaultOptions['limit_page'] = (int) $this->defaultOptions['limit_page'];
            $this->defaultOptions['limit_per_page'] = (int) $this->defaultOptions['limit_per_page'];
            $limit_offset  =  ( $this->defaultOptions['limit_page'] -1) * ($this->defaultOptions['limit_per_page']);
            $limit_amount = $this->defaultOptions['limit_per_page'];
        }  else {      
            $limit_offset = $this->defaultOptions['limit_offset'];
            //$limit_amount = $this->defaultOptions['limit_per_page'];
            $limit_amount = false;
            if ($limit_offset !==false || $limit_offset !== 'false' )  {
                if ( !(is_integer($limit_offset) || ctype_digit($limit_offset)) || $limit_offset < 0) {
                    $limit_offset =0;
                }
            }
            if ($limit_amount === null || empty($limit_amount) || $limit_amount === 'false') {
                $limit_amount = false;
            }
            if ($limit_amount !== false) {
                if ( !(is_integer($limit_amount) || ctype_digit($limit_amount)) || $limit_amount < 1) {
                    $limit_amount = 100;
                }
            }
        }
        $sort_order = explode(',',$this->defaultOptions['sort_order'] . '');        
        $fieldData = $this->getDisplayFieldsData();
        $fields = array();
        $groups = array();
        $aggregates = array();
        $aggregate_fields = array();
        $has_total = false;

       
        $cols_in_report = I2CE_CustomReport::getColumnsInReportTable($this->config->report);        
        foreach ($fieldData as $field=>$data) {
            if ($field == 'total') {
                $has_total = true;
                continue;
            }
            list($form_form,$form_field, $aggregate) =  array_pad(explode( '+', $field ),3,''); //function fields have blank $form_form
            if (!in_array($form_form . '+' . $form_field,$cols_in_report)) {
                I2CE::raiseError("Skipping field $field b/c it is not in report " . $this->config->report);
                continue;
            }
            $fields[] = "$form_form+$form_field";
            if ($form_form)   {
                //it is not a function field.  check if there is a link
                if ($data['link'] && $data['link_append']) {
                    $groups[$data['link_append']] = $data['link_append'];
                }
            }
            $reportfield = "$form_form+$form_field";
            switch ($aggregate) {
            case 'sum':
                $aggregates[$reportfield] = " SUM(`$reportfield`) AS `$field`";
                $aggregate_fields[$reportfield] = "$field";
                break;
            case 'maximum':
                $aggregates[$reportfield] = " MAX(`$reportfield`) AS `$field`";
                $aggregate_fields[$reportfield] = "$field";
                break;
            case 'minmimum':
                $aggregates[$reportfield] = " MIN(`$reportfield`) AS `$field`";
                $aggregate_fields[$reportfield] = "$field";
                break;
            case 'average':
                $aggregates[$reportfield] = " AVG(`$reportfield`) AS `$field`";
                $aggregate_fields[$reportfield] = "$field";
                break;
            case 'count':
                $aggregates[$reportfield] = " COUNT(`$reportfield`) AS `$field`";
                $aggregate_fields[$reportfield] = "$field";
                break;
            case 'count_distinct':
                $aggregates[$reportfield] = " COUNT(DISTINCT `$reportfield`) AS `$field`";
                $aggregate_fields[$reportfield] = "$field";
                break;
            default: //'none'
                $groups[$reportfield] = $reportfield;
            }
        }
        if (count($fields) == 0 && !$has_total) {
            I2CE::raiseError("No fields to get results from");
            return false;
        }
        $sorts = array();
        foreach ($sort_order as $formfield) {
            if (strlen($formfield) == 0) {
                continue;
            }
            if ($formfield[0] == '-') {
                $sort_field = substr($formfield,1);
                $sort_postfix = ' DESC';
            } else {
                $sort_field = $formfield;
                $sort_postfix = '';
            }
            if ($sort_field != 'total' && !in_array($sort_field,$fields)) { //make sure we don't have junk
                continue;
            }
            if (array_key_exists($sort_field,$aggregate_fields)) {
                $sort_field = $aggregate_fields[$sort_field];
            }
            $sorts[$sort_field] = '`' . $sort_field . '`' . $sort_postfix;
        }
        $order_by = '';
        if (count($sorts) > 0) {
            $order_by = ' ORDER BY ' . implode(',',$sorts);
        }
        $group_by = '';
        if (count($aggregates) == 0) { //no aggregate fields
            if ($has_total) {
                $select_fields = array();
                foreach ($groups as $group) {
                    $select_fields[] = '`' . mysql_real_escape_string($group) . '`';
                }
                if (count($select_fields) > 0) {
                    $group_by = ' GROUP BY ' . implode(',',$select_fields);
                }
                $select_fields[] ='COUNT(*) AS `total`';
            }else{
                if (count($groups) == 0) {
                    I2CE::raiseError("Report view {$this->view} has no viewable fields");
                    return false;
                }
                $select_fields = array();
                foreach ($groups as $group) {
                    $select_fields[] = '`' . mysql_real_escape_string($group) . '`';
                }
            }
        } else {            
            $select_fields = $aggregates;
            $group_bys = array();
            foreach ($groups as $group) {
                $group = '`' . mysql_real_escape_string($group) . '`';
                $select_fields[] = $group;
                $group_bys[]  = $group;
            }
            if ($has_total) {
                $group_by = ' GROUP BY ' . implode(',',$select_fields);
                $select_fields[] ='COUNT(*) AS `total`';                
            } else {
                if (count($group_bys) > 0) {
                    $group_by = ' GROUP BY ' . implode(',',$group_bys);
                }            
            }
        }                     
        $select_field_list = implode(',',$select_fields);
        if ($limit_offset !== false && $limit_amount !== false) {
            $limit = " LIMIT $limit_amount OFFSET $limit_offset ";
            $this->row_start = $limit_offset;
            $this->row_amount = $limit_offset;
        } else {
            $limit  = '';
            $this->row_start = false;
            $this->row_amount = false;
        }
        if (array_key_exists('nested_limits',$this->defaultOptions)) {
            $where = $this->processWhere($this->defaultOptions['nested_limits']);
        } else {
            $where = '';
        }
        $qry = "SELECT SQL_CALC_FOUND_ROWS $select_field_list FROM " . 
            I2CE_CustomReport::getCachedTableName($this->config->report) .
            ' ' . $where . ' ' . $group_by . ' ' . $order_by . ' ' . $limit;
        $db = MDB2::singleton();
        $res = $db->query($qry);
        if (I2CE::pearError($res,"Could not get results")) {
            return false;
        } 
        $num_rows = $db->queryRow( "SELECT FOUND_ROWS() AS num_rows" );
        if (I2CE::pearError($num_rows,"Could not get total number of results")) {
            $num_rows  = false;
        } else {
            $num_rows = (int) $num_rows->num_rows;
        }
        return array('results'=>$res,'num_results'=>$num_rows, 'has_total' => $has_total );        
    }

    /**
     * The starting row (if any) for limiting the results. If not used, it is false
     * @var procted mixed $row_start
     */
    protected $row_start;
    /**
     * The number of rows (if any) for limiting the results. If not used, it is false
     * @var procted mixed $row_amount
     */
    protected $row_amount;

    /**
     * Maps any mapped  fields in a results.
     * @param array $result A an array with inicies of the form "$reportform+$field(+$aggregate)"  with unmapped value
     * @return array the mapped results
     */
    protected function mapResults($result) {
        $ret = array();
        foreach ($result as $reportformfield => $value) {
            $ret[$reportformfield] = $this->getMappedValue($reportformfield,$value);
        }       
        return $ret;
    }
    

    /**
     * Gets the mapped value for a report form field.
     * @param string $reportformfield or a report form field or a form field of the form "$reportform+$field" or "$reportform+$field+$aggreate" or "$form+field"
     * @param value
     * @parmam boolean $isReportForm Defaults to true.  If true we a sending a report form field
     * If false, we are working with a form field.
     * @returns string the origina value of  on failure, the mapped value on success
     */
    protected function getMappedValue($reportformfield,$value,$isReportForm=true) {
        $this->setupMappedValues($reportformfield,$isReportForm); //makes sure the map is setup
        if ($isReportForm) {
            $this->mapReportFormField($reportformfield);
            $formfield = $this->fieldMaps[$reportformfield];
        } else {
            $formfield = $reportformfield;
        }        
        if ( !$formfield ) {
            list($form,$field,$aggregate) = array_pad(explode('+',$reportformfield),3,'');
            if ( $form == "" ) {
                // This may be a function
                $formfield_type = null;
                if ( $this->relationshipConfig->setIfIsSet( $formfield_type, 
                            "reporting_functions/$field/formfield" ) ) {
                    $formfield_class = null;
                    if ( I2CE::getConfig()->setIfIsSet( $formfield_class, 
                                "/modules/forms/FORMFIELD/$formfield_type" ) ) {
                        if ( is_subclass_of( $formfield_class, "I2CE_FormField_MAPPED" ) ) {
                            $options = array( "meta" => array() );
                            if ( $this->relationshipConfig->setIfIsSet( $select_forms, "reporting_functions/$field/select_forms" ) && $select_forms ) {
                                $options["meta"]["form"] = 
                                    explode( ",", $select_forms );
                            }
                            if ( $this->relationshipConfig->setIfIsSet( $linked_fields, "reporting_functions/$field/link_fields" ) && $linked_fields ) {
                                $options["meta"]["display"] =
                                    array( "default" => 
                                            array( "fields" => 
                                                $linked_fields )
                                         );
                            }
    
                            $ff_obj = new $formfield_class( $field, $options );
                            $ff_obj->setFromDB( $value );
                            $ff_value = $ff_obj->getDisplayValue();
                            return $ff_value;
                        }
                    }
                }
            }
            return $value;
        }
        if (is_string($formfield) && array_key_exists($formfield,$this->mappedValues)) {
            $map_value = false;
            foreach ($this->mappedValues[$formfield] as $data) {
                if (array_key_exists('display',$data) && $data['display'] == $value) {
                    $map_value = $data['value'];
                    break;
                }
            }
            if ($map_value !== false) {
                return $map_value;
            }
        }
        list($form,$field) = array_pad(explode('+',$formfield),2,'');
        if (!array_key_exists($form,$this->formObjs) || !$this->formObjs[$form] instanceof I2CE_Form) {
            return $value;
        }
        $fieldObj = $this->formObjs[$form]->getField($field);
        if (!$fieldObj instanceof I2CE_FormField) {
            return $value;
        }
        $fieldObj->setFromDB($value);
        return $fieldObj->getDisplayValue();
    }
    /**
     * Stores the array of mapped value for a formfield in  {@var $formfield}
     * @parmam boolean $isReportForm Defaults to true.  If true we a sending a report form field
     * If false, we are working with a form field.
     * @param string $reportformfield
     */
    protected function setupMappedValues($reportformfield,$isReportForm = true) {
        if ($isReportForm) {
            $this->mapReportFormField($reportformfield);
            $formfield = $this->fieldMaps[$reportformfield];
            if(!$this->isMapped($reportformfield)) {
                return;
            }
        } else {
            $formfield = $reportformfield;
        }
        if ($formfield === false) {
            return;
        }
        if (array_key_exists($formfield,$this->mappedValues)) {
            //we have already gotten the mapped values
            return;
        }
        //we need to get the mapped values
        list($formObj,$fieldObj) = $this->getFormFieldObjects($reportformfield,$isReportForm);
        if (!$formObj instanceof I2CE_Form) { //we are dealing with a funcion field which is not mapped so return
            return; 
        }
        $mappedValues = array();
        if ($fieldObj instanceof I2CE_FormField) {
            $mappedValues = $fieldObj->getMapOptions();
        }
        if (!is_array($mappedValues)) { //just to be sure
            $mappedValues = array();
        }
        $this->mappedValues[$formfield] = $mappedValues;
    }


    /**
     * Checks to see if a form field is mapped
     * @param string $reportformfield of the form "$reportform+$fiedld(+$aggregate)"
     */
    protected function isMapped($reportformfield) {
        if (!array_key_exists ($reportformfield,$this->mappedFields)) {
            //we have not check this form field yet.
            list($formObj,$fieldObj) = $this->getFormFieldObjects($reportformfield);
            if (!$formObj instanceof I2CE_Form) {
                $ismap  = false;
            } else if (!$fieldObj instanceof I2CE_FormField) {
                $ismap  = false;
            } else {
                $ismap = $fieldObj instanceof I2CE_FormField_MAPPED;
            }
            $this->mappedFields[$reportformfield] = $ismap;
        }
        return $this->mappedFields[$reportformfield];
    }


    /**
     * Sets the maps
     * @param string $reportformfield
     */
    protected function mapReportFormField($reportformfield) {
        if (array_key_exists($reportformfield,$this->formMaps)) {
            return;
        }
        list($reportform,$field,$aggregate) = array_pad(explode('+',$reportformfield),3,'');
        if (strlen($reportform) == 0) { //this is a function field
            $reportform = false;
            $form = false; 
        } elseif ( !$field && $reportform == "total" ) {
            $this->fieldMaps[$reportform] = false;
            return;
        } else {            
            $form  = $this->reportObj->getFormRelationship()->getForm($reportform);
            if (!is_string($form) || strlen($form)== 0) {
                $form = false;
            }
        }
        $this->formMaps[$reportformfield] = $form;
        $this->formMaps["$reportform+$field"] = $form;
        $this->formMaps[$reportform] = $form;
        if ($form == false) {
            $this->fieldMaps[$reportformfield] = false;
            $this->fieldMaps["$reportform+field"] = false;
            $this->fieldMaps[$reportform] = false;
        } else {
            $this->fieldMaps[$reportformfield] = "$form+$field";
            $this->fieldMaps["$reportform+field"] = "$form+$field";
            $this->fieldMaps[$reportform] = "$form+$field";
        }
    }
    

    /**
     * Get the form and field objects associated to a formfield string.  
     * @param string $reportformfield of the re[prt form "$form" "$form+$field"  or "$form+$field+$aggregate"
     * @parmam boolean $isReportForm Defaults to true.  If true we a sending a report form field
     * If false, we are working with a form field.
     * @returns array.  Index 0 is an I2CE_Form (on success, false on failure), Index 1 is an I2CE_FormField on succces, false on failure
     */
    protected function getFormFieldObjects($reportformfield,$isReportForm = true) {
        $ret = array(0=>false,1=>false);
        if ($isReportForm) {
            $this->mapReportFormField($reportformfield);
            if (!is_string($this->fieldMaps[$reportformfield])) {
                return $ret;
            }
            $formfield = $this->fieldMaps[$reportformfield];
        } else {
            $formfield = $reportformfield;
        }
        list($form,$field,$aggregate) = array_pad(explode('+',$formfield),3,'');
        if (!$form) { //it is a function field so there are no form/field objects associated to it/
            return $ret;
        }
        if (!array_key_exists($form,$this->formObjs)) {
            $factory  = I2CE_FormFactory::instance();
            $this->formObjs[$form] = $factory->createForm($form);
            if (!$this->formObjs[$form] instanceof I2CE_Form) {
                $this->formObjs[$form] = false;
            }
        }
        $ret[0] = $this->formObjs[$form];
        if (!$ret[0] instanceof I2CE_Form) {
            return $ret;
        }
        $ret[1] = $ret[0]->getField($field);
        if (!$ret[1] instanceof I2CE_FormField) {
            $ret[1] = false;
        }
        return $ret;
    }




    /**
     * Process the fields  to make limits based on their limiting values
     * @param array $limits an array defining limiting values for particular fields
     * @returns string
     */
    protected function processWhere($limitValues) {        
        if (!is_array($limitValues) || count($limitValues) == 0) {
            return '';
        }
        $wheres = array();
        $factory = I2CE_FormFactory::instance();
        $t_excludes =   I2CE::getConfig()->modules->CustomReports->limit_excludes->displayed->getAsArray();
        $excludes = array();
        foreach ($t_excludes as $exclude) {
            $excludes[$exclude] = true;
        }
        foreach ($limitValues as $reportformfield=>$limit_styles) {
            if (!is_array($limit_styles)) {                
                continue;
            }
            $limit_styles = array_diff_key($limit_styles,$excludes);
            if (count($limit_styles) == 0) {
                continue;
            }            
            list($reportform,$field) = array_pad(explode('+',$reportformfield),2,'');
            if (!$reportform) { //function fields have no limits.
                $where = trim($this->processWhereByFunction($field,$limit_styles));
            } else {
                list($formObj,$fieldObj)= $this->getFormFieldObjects($reportformfield);
                if (!$formObj instanceof I2CE_Form) {
                    continue;
                }            
                $where = trim($this->processWhereByField($reportform,$field,$formObj,$limit_styles));
            }
            if (strlen($where) >0) {
                $wheres[] = '(' . $where . ')';
            }
        }
        $where = '';
        if (count($wheres) > 0) {
            $where = ' WHERE (' . implode(' AND ' , $wheres) . ') ';
        }
        return $where;
    }

    
    /**
     * @param string $form  the (report) form  
     * @param string $field the field
     * param I2CE_Form $formObj the instantiation of the form that the report form references 
     * @param array $limitStyles the limit values for this formfield indexed by limit type
     */
    protected function processWhereByField($form,$field,$formObj,$limitStyles) {        
        if (!is_array($limitStyles) || count($limitStyles) == 0 ) {
            return '';
        }
        $config = $this->reportConfig->traverse("reporting_forms/$form/fields/$field/limits",false,false);
        if (!$config instanceof I2CE_MagicDataNode) {
            return '';
        }
        $wheres = array();
        $callback = create_function('$form,$field', "return '`$form+$field`';"); //weird, but i do want $form and $field substituted now and now later.
        foreach ($limitStyles  as $limitStyle=>$values) {
            if (!isset($config->$limitStyle)) {
                continue;
            }
            if (!isset($config->$limitStyle->enabled) || !$config->$limitStyle->enabled) {
                continue;
            }
            if (!is_array($values) || count($values)==0) {
                continue;
            }
            $where =  trim($formObj->generateLimit(array('field'=>$field,'style'=>$limitStyle,'data'=>$values), $callback));
            if (strlen($where) >0) {
                $wheres[] = '(' . $where . ')';
            }
        }
        $where = '';
        if (count($wheres) >0) {
            $where = '(' . implode ( ' AND ' , $wheres) . ')';
        }
        return $where;        
    }


    
    /**
     * @param string $function
     * @param array $limitStyles the limit values for this formfield indexed by limit type
     */
    protected function processWhereByFunction($function,$limitStyles) {        
        if (!is_array($limitStyles) || count($limitStyles) == 0 ) {
            return '';
        }
        $config = $this->reportConfig->traverse("reporting_functions/$function/limits",false,false);
        if (!$config instanceof I2CE_MagicDataNode) {
            return '';
        }
        $wheres = array();
        foreach ($limitStyles  as $limitStyle=>$values) {
            if (!isset($config->$limitStyle)) {
                continue;
            }
            if (!isset($config->$limitStyle->enabled) || !$config->$limitStyle->enabled) {
                continue;
            }
            if (!is_array($values) || count($values)==0) {
                continue;
            }
            $repFunction = $this->getSwissReportFunction($function); //get rid of the leading +
            if (!$repFunction instanceof I2CE_Swiss_CustomReports_Report_ReportingFunction) {
                continue;
            }
            $fieldObj = $repFunction->getFieldObj();
            if (!$fieldObj instanceof I2CE_FormField) {
                continue;
            }
            $where =  trim($fieldObj->generateLimit(array('style'=>$limitStyle,'data'=>$values), '`+' . $function .'`'));
            if (strlen($where) >0) {
                $wheres[] = '(' . $where . ')';
            }
        }
        $where = '';
        if (count($wheres) >0) {
            $where = '(' . implode ( ' AND ' , $wheres) . ')';
        }
        return $where;        
    }





        
    /**
     * returns an array of the default display options.  Default options are read and overriddenb
     * in the following order:
     * /modules/CustomReports/displays/$display/display_options
     * /modules/CustomReports/relationships/display_options/$display
     * /modules/CustomReports/reports/display_options/$display
     * /modules/CustomReports/reportViews/$view/display_options/$display
     * Finally any options that have a key in $get are replaced by that value
     * @param array $get
     * @param array $options. Default to the empty array.  The options that we want to be already set before we start goinng through
     * @returns mixed array or false on failure
     */
    protected function getDefaultOptions($get,$options = array() ) {
        //make sure we can overide the following options which may come from the $get varaible.
        $make_exist = array('limit_paginated','limit_page','limit_per_page',
                            'limit_offset','limit_amount','sort_order','display_order',
                            'total','save_options_as_default_view');
        foreach ($make_exist as $key) {
            if (!array_key_exists($key,$options) ) {
                $options[$key] = null;
            }
        }    
        $check = array(
            "/modules/CustomReports/default_display_options"=>I2CE::getConfig(),
            "/modules/CustomReports/displays/{$this->display}/display_options"=>I2CE::getConfig(),
            'default_display_options'=>$this->relationshipConfig,
            "display_options/{$this->display}"=>$this->relationshipConfig,
            'default_display_options'=>$this->reportConfig,
            "display_options/{$this->display}"=>$this->reportConfig,
            'default_display_options'=>$this->config,
            "display_options/{$this->display}"=>$this->config);

        // I don't know if this is the right way to do this or not.
        // I'm guessing probably not, but for a new report I added no other options were set
        // so maybe the edit needs to change to save the data elsewhere.
        $this->config->setIfIsSet( $options['total'], "total" );
        
        foreach($check as $k=>$node) {
            $t_options = array();
            if ( $node->setIfIsSet($t_options,$k,true)) {
                I2CE_Util::merge_recursive($options,$t_options);
            }
        }
        //now look at the get variables
        I2CE_Util::merge_recursive($options,$get,true,false); //add new keys but not the empty ones
        //finally load some limiting stuff from get variables that not have been set in the default options above.
        //and convert it over to the nested
        if (!array_key_exists('nested_limits',$options) || !is_array($options['nested_limits'])) {
            $options['nested_limits'] = array();
        }
        if (array_key_exists('limits',$get) && is_array($get['limits'])) {
            I2CE_Util::merge_recursive($options['nested_limits'],$get['limits'],true,false);  //add new keys but not the empty ones
        }
        return $options;
    } 





    /**
     * Adds any report display controls that can be added for this view.
     * @param DOMNode $conentNode
     * @param mixed $controls.  If null (default), we display all the report controsl.  If string or an array of string, we only display the indicated controls
     * @returns boolean $true on success
     */
    protected function displayReportControls($contentNode, $controls=null) {
        $displayConfig = I2CE::getConfig()->modules->CustomReports->displays;
        if (is_string($controls)) {
            $controls= array($controls);
        }
        $displays = $displayConfig->getKeys(); 
        if (is_array($controls)) {
            $displays = array_intersect($displays,$controls);
        }
        if (count($displays)== 0) {
            $displays[] = 'Default';
        }
        foreach ($displays as $display ) {
            if ($display != $this->display ) {
                if (!isset($displayConfig->$display->class) || ! ($displayConfig->$display->class)) {
                    continue;
                }
                $displayClass = $displayConfig->$display->class;
                if (!I2CE::getFileSearch()->search('CLASSES',$displayClass . '.php')) {
                     //this is so we don't clutter up the log with a warning if the class file is not there.
                    continue;
                }
                if (!class_exists($displayClass)) {  //skip if something is wrong with the class as define by the class file (will generate a warning)
                    //I2CE::raiseError("Cannot find class $displayClass as a subclass of I2CE_CustomReport_Display");
                    continue;
                }
                if ( !is_subclass_of($displayClass,'I2CE_CustomReport_Display')) {
                    continue;
                }

                try {
                    $displayObj =  new $displayClass($this->page,$this->view);
                }
                catch (Exception $e) {
                    continue;
                }
            } else {
                $displayObj = $this;
            }
            $controlNode = $this->template->createElement('span',array('class'=>"CustomReport_control",'id'=>"CustomReport_controls_$display"));            
            $contentNode->appendChild($controlNode);
            $displayObj->displayReportControl($controlNode);
        }
        return true;
    }
    
    protected function getBasePage() {
        $module = $this->page->module();
        if ($module == 'I2CE') {
            $base_page = $this->page->page();
        } else {
            $base_page = $module .'/' . $this->page->page();
        }
        return $base_page;
    }

    /**
     * Adds any controls for this display to the content node
     * @param DOMNode $contentNode 
     */
    protected function displayReportControl($contentNode) {
        $fileSearch = I2CE::getFileSearch();
        if ($fileSearch->search('CSS', "customReports_display_control_{$this->display}.css")) {
            $this->template->addHeaderLink("customReports_display_control_{$this->display}.css");
        }
        if ($fileSearch->search('TEMPLATES', "customReports_display_control_{$this->display}.html")) {
            $controlNode = $this->template->appendFileByNode("customReports_display_control_{$this->display}.html",'div',$contentNode);
            if (!$controlNode instanceof DOMNode ){
                I2CE::raiseError("Could not find template for ({$this->display}) control");
                return false;
            }
        } else {
            return true;
        }
        $base_page = $this->getBasePage();
        //$submitNodes = $this->template->query('.//input[@type="submit" and @name="' .  $this->display .  '_submit"]',$controlNode);
        //Some nodes were changed to use id and got rid of name, so check for both
        $submitNodes = $this->template->query('.//input[@type="submit" and ( @name="' .  $this->display .  '_submit" or @id="' . $this->display . '_submit" )]',$controlNode);
        for ($i=0; $i< $submitNodes->length; $i++) {
            $submitNode = $submitNodes->item($i);
            $class = $submitNode->getAttribute('class') . " method=post action=index.php/$base_page/{$this->view}/{$this->display}";            
            $submitNode->setAttribute('class',$class);            
        }
        $submitNodes = $this->template->query('.//input[@type="submit" and @name="' .  $this->display .  '_saveOptions"]',$controlNode);
        for ($i=0; $i< $submitNodes->length; $i++) {
            $submitNode = $submitNodes->item($i);
            $class = $submitNode->getAttribute('class') . " method=post action=index.php/$base_page/saveOptions/{$this->view}/{$this->display}";            
            $submitNode->setAttribute('class',$class);            
        }
        return true;
    }
    



    protected  $reportViewsFactory;
    protected  function getReportViewsFactory() {
        if ($this->reportViewsFactory instanceof I2CE_SwissMagicFactory) {
            return $this->reportViewsFactory;
        }
        $repViewsConfig = $this->config->traverse('..');
        if (!$repViewsConfig instanceof I2CE_MagicDataNode) {
            I2CE::raiseError("Bad magic data path to report views");
            return false;
        }
        $init_options = array(
            'root_url_postfix'=>'reportViews',
            'root_path'=>$repViewsConfig->getPath(false),
            'root_type'=>'CustomReports_ReportViews');
        try {
            $swiss_factory = new I2CE_SwissMagicFactory($this->page,$init_options);
        } catch (Exception $e) {
            I2CE::raiseError("Could not create swissmagic for reportViews:" . $e->getMessage());
            return false;
        }
        try {
            $swiss_factory->setRootSwiss();
        } catch (Exception $e) {
            I2CE::raiseError("Could not create root swissmagic for relationships:" . $e->getMessage());
            return false;
        }      
        $this->reportViewsFactory = $swiss_factory;
        return $this->reportViewsFactory;
    }


    protected function getSwissReportView() {
        $reportViewsFactory = $this->getReportViewsFactory();
        if (!$reportViewsFactory instanceof I2CE_SwissMagicFactory) {
            return false;
        }
        return $reportViewsFactory->getSwiss('/' . $this->view);
    }

    protected function getSwissReport() {
        $reportView = $this->getSwissReportView();
        if(!$reportView instanceof I2CE_Swiss_CustomReports_ReportView) {
            return false;
        }
        return $reportView->getSwissReport($this->config->report);
    }
    

    protected function getSwissReportFunction($func) {
        $swissReport = $this->getSwissReport();
        if (!$swissReport instanceof I2CE_Swiss_CustomReports_Report) {
            return false;
        }
        $funcs = $swissReport->getChild('reporting_functions');
        if (!$funcs instanceof I2CE_Swiss_CustomReports_Report_ReportingFunctions) {
            return false;
        }
        return $funcs->getChild($func);
    }
    


    /**
     * Displays any report limits in the content node
     * @param DOMNode $contentNode
     */
    protected function displayReportLimits($contentNode) {       
        if (array_key_exists('nested_limits',$this->defaultOptions) && is_array($this->defaultOptions['nested_limits'])) {
            $limitValues =$this->defaultOptions['nested_limits'];
        } else {
            $limitValues = array();
        }
        $excludes = I2CE::getConfig()->modules->CustomReports->limit_excludes->displayed->getAsArray();
        $swissRel = false;
        foreach ($this->reportConfig->reporting_functions as $func=>$funcConfig) {
            if (!$funcConfig instanceof I2CE_MagicDataNode || !$funcConfig->is_parent('limits')) {
                continue;
            }
            if (array_key_exists('+'.$func,$limitValues) ) {
                $funcLimitValues = $limitValues['+'.$func];
            } else {
                $funcLimitValues = array();
            }
            foreach ($funcConfig->limits as $limit=>$limitConfig) {
                if (!$limitConfig instanceof I2CE_MagicDataNode 
                    ||!$limitConfig->is_scalar('enabled')
                    ||!$limitConfig->enabled) {
                    continue;
                }
                if (array_key_exists($limit,$funcLimitValues) && is_array($funcLimitValues[$limit] )) {
                    $limitLimitValues = $funcLimitValues[$limit];
                } else {
                    $limitLimitValues = array();
                }
                $repFunction = $this->getSwissReportFunction($func);
                if (!$repFunction instanceof I2CE_Swiss_CustomReports_Report_ReportingFunction) {
                    continue;
                }
                $fieldObj = $repFunction->getFieldObj();
                if (!$fieldObj instanceof I2CE_FormField) {
                    continue;
                }
                $method = 'processLimitMenu_' . $limit;                                        
                $data = $fieldObj->$method($limitLimitValues,false);
                $method = 'getLimitMenu_' . $limit;
                $node = $fieldObj->$method($this->template,"limits:+$func:$limit",$data);
                if ( !$node instanceof DOMNode) {
                    continue;
                }                       
                $this->displayReportLimit($contentNode,$node,$limitConfig);                
            }            
        }


        $reportforms  = $this->reportConfig->reporting_forms->getKeys();
        foreach ($reportforms as $reportform) {
            $fields = $this->reportConfig->getKeys("reporting_forms/$reportform/fields");
            foreach ($fields as $field) {
                $reportformfield = "$reportform+$field";
                if (array_key_exists($reportformfield,$limitValues) ) {
                    $fieldLimitValues = $limitValues[$reportformfield];
                } else {
                    $fieldLimitValues = array();
                }
                list ($formObj,$fieldObj) = $this->getFormFieldObjects("$reportform+$field");
                if (!$formObj instanceof I2CE_Form) {
                    continue 2;
                }
                if (!$fieldObj instanceof I2CE_FormField) {
                    continue;
                }
                $limits = $formObj->getLimitStyles($field);
                $limits = array_diff($limits,$excludes);
                $limit_default = 'default';
                $this->reportConfig->setIfIsSet($limit_default,"reporting_forms/$reportform/fields/$field/limit_default");
                foreach ($limits as $limit) {
                    if (!$this->reportConfig->pathExists("reporting_forms/$reportform/fields/$field/limits/$limit")) {
                        continue;
                    }                                         
                    $limitConfig = $this->reportConfig->traverse("reporting_forms/$reportform/fields/$field/limits/$limit");
                    if (!isset($limitConfig->enabled) || !$limitConfig->enabled) {
                        continue;
                    }
                    if (array_key_exists($limit,$fieldLimitValues) && is_array($fieldLimitValues[$limit] )) {
                        $limitLimitValues = $fieldLimitValues[$limit];
                    } else {
                        $limitLimitValues = array();
                    }
                    //now we should have a valid limit
                    $method = 'processLimitMenu_' . $limit;                                        
                    $data = $fieldObj->$method($limitLimitValues,false);
                    $method = 'getLimitMenu_' . $limit;
                    $node = $fieldObj->$method($this->template,"limits:$reportformfield:$limit",$data,$limit_default);
                    if ( !$node instanceof DOMNode) {
                        continue;
                    }        
                    $this->displayReportLimit($contentNode,$node,$limitConfig);
                }
            }
        }
        return true;
    }
    
    



    /**
     * Adds the report limit node to its content node
     * @param DOMNode $contentNode
     * @param DOMNode $limitnode
     * @param I2CE_MagicDataNode $limitConfig
     */
    protected function displayReportLimit($contentNode,$limitNode,$limitConfig) {
        $trNode = $this->template->createElement('tr');
        $contentNode->appendChild($trNode);
        if (!$limitConfig->is_scalar('header')) {
            $header = $limitConfig->getName();
        } else {
            $header = $limitConfig->header;
        }
        $tdNode = $this->template->createElement('td',array(),$header);
        $trNode->appendChild($tdNode);

        $tdNode = $this->template->createElement('td');
        $tdNode->appendChild($limitNode);
        $trNode->appendChild($tdNode);
    }


    
    public function saveDefaultView() {
        if (!array_key_exists( 'save_options_as_default_view',$this->defaultOptions)) {
            return;
        }
        if (!$this->defaultOptions['save_options_as_default_view']) {
            return;
        }
        unset($this->defaultOptions['save_options_as_default_view']);
        if (!is_string($this->display) || strlen($this->display) ==0) {
            return;
        }
        $this->config->default_display = $this->display;
        $this->saveDisplayOptions();
    }

    public function saveDisplayOptions() {
        if (!is_string($this->display) || strlen($this->display) ==0) {
            return false;
        }
        $this->config->display_options->{$this->display}->erase();       
        $this->config->display_options->{$this->display} = $this->defaultOptions;
        return true;
    }


    /**
     * Display the report
     * @param DOMNode $contentNode The DOM node we wish to display into. If null, we do not do any of the DOM processing stuff, do
     * not call the report display controls, limits etc. It will however still call processResults with a DOMNode of null
     * @param boolean $processResults Defaults to true meaning we run through the results.  If false, we do not process results.
     * @param mixed $controls.  If null (default), we display all the report controsl.  If string or an array of string, we only display the indicated controls
     * @returns boolean. true on sucess
     */
    public function display($contentNode,$processResults = true, $controls = null) {
        $this->saveDefaultView();
        if ($contentNode instanceof DOMNOde) {
            $baseHTML = "customReports_display_{$this->display}_base.html";
            $this->displayConfig->setIfIsSet($baseHTML,'base_html');
            $baseNode = $this->template->appendFileByNode( $baseHTML, 'div',$contentNode );        
            if (!$baseNode instanceof DOMNode) {
                I2CE::raiseError("Unable to add {$this->display} display base template");
                return false;
            }        
            if (!$this->displayMetaInfo($baseNode)) {
                I2CE::raiseError("Could not process results");
                return false;
            }
            $reportLimitsNode = $this->template->getElementById('report_limits', $baseNode);
            if (!$reportLimitsNode instanceof DOMNode ){
                I2CE::raiseError("Don't know where to add report limits");
                return false;
            }
            if (!$this->displayReportLimits($reportLimitsNode)) {
                return false;
            }
            $reportControlsNode = $this->template->getElementById('report_controls', $baseNode);
            if (!$reportControlsNode instanceof DOMNode ){
                I2CE::raiseError("Don't know where to add report controls");
                return false;
            }
            if (!$this->displayReportControls($reportControlsNode, $controls)) {
                return false;
            }
            $resultsNode = $this->template->getElementById('report_results', $baseNode);
            if (!$resultsNode instanceof DOMNode ){
                I2CE::raiseError("Don't know where to add report results");
                return false;
            }
        } else {
            $resultsNode = null;
        }
        if ($processResults) {
            $data = $this->getResults();
            $this->template->setDisplayData('show_results','1');
            if (!$this->processResults($data,$resultsNode)) {                
                return false;
            }            
        } else {
            $data = false;
            $this->template->setDisplayData('show_results','');
        }
        return true;
    }

    /**
     * Process results
     * @param array $results_data an array of results.  indices are 'restults' and MDB2 Buffered result and 'num_results' the
     * number of results.  (these values may be false on failure)
     * @param DOMNode $contentNode.  Default to null a node to append the results onto
     */
    protected function processResults($results_data,$contentNode=null) {
        //$results_data['num_results']  is the total number of results or false
        if ($results_data['results'] == false) {
            I2CE::raiseError("No results");
            return false;
        }
        if ($this->row_start === false) {
            $row_num = 0;
        } else {
            $row_num = $this->row_start;
        }        
        while ($row = $results_data['results']->fetchRow()) {
            if (  PEAR::isError( $row ) ) {
                break;
            }
            if (!$this->processResultRow($row,$row_num,$contentNode)) {
                $results_data['results']->free();
                return false;
            }
            $row_num++;
        }
        $results_data['results']->free();
        return true;
    }

    
    /**
     * Process a result row.
     * @param array $row
     * @param int $row_num The current row number when processing results.  If there was a result limit, it starts the count from the beginning of the
     * result offset.  Othwerwise, it starts counting form zero.
     * @param DOMNode $contentNode. Default to null. A node to append the result onto
     * @returns boolean. True on success
     */
    protected function processResultRow($row,$row_num,$contentNode=null) {
        return true;
    }
   

    protected function displayMetaInfo($contentNode) {
        if (!$this->template instanceof I2CE_Template) { //this is not a web page -- maybe its xml i.e. only an instanceof I2CE_TemplateMeister
            I2CE::raiseError("Unexpected template type " . get_class($this->template));
            return false;
        }
        $form = $this->template->query(".//form[@id='limit_form']",$contentNode);
        if ($form->length == 1) {
            //$form = $form->item(0)->setAttribute('action',$this->getBasePage() . "/show/{$this->view}/{$this->display}");            
            // The above was causing problems with having /show/show in the URL so I took it out.
            $form = $form->item(0)->setAttribute('action',$this->getBasePage() . "/{$this->view}/{$this->display}");            
        }   else {
            I2CE::raiseError("Could not form node");
            return false;
        }            
        $formWormOptions = array('optionsMenuPositionVert'=>'mouse_above');
        $this->template->addFormWorm('limit_form',$formWormOptions);
        $fileSearch = I2CE::getFileSearch();
        if ($fileSearch->search('CSS', "customReports_display.css")) {
            $this->template->addHeaderLink("customReports_display.css");
        }
        if ($fileSearch->search('CSS', "customReports_display_{$this->display}.css")) {
            $this->template->addHeaderLink("customReports_display_{$this->display}.css");
        }
        $this->template->setDisplayDataImmediate('report_view_display_name',$this->config->display_name,$contentNode);
        $this->template->setDisplayDataImmediate('report_view_description',$this->config->description,$contentNode);
        $this->template->setDisplayDataImmediate('limit_paginated',$this->defaultOptions['limit_paginated'],$contentNode);
        $this->template->setDisplayDataImmediate('limit_page',$this->defaultOptions['limit_page'],$contentNode);
        $this->template->setDisplayDataImmediate('limit_per_page',$this->defaultOptions['limit_per_page'],$contentNode);        
        $this->template->setDisplayDataImmediate('limit_offset',$this->defaultOptions['limit_offset'],$contentNode);
        $this->template->setDisplayDataImmediate('limit_amount',$this->defaultOptions['limit_amount'],$contentNode);        
        $this->template->setDisplayDataImmediate('sort_order',$this->defaultOptions['sort_order'],$contentNode);
        $time = I2CE_CustomReport::getLastGenerationTime($this->config->report);
        if ($time) {
            $cache_msg = "Report was generated at: " . strftime("%c",$time);
        } else {
            $cache_msg = "Report has not been generated";
        }
        $this->template->setDisplayDataImmediate('report_time',$cache_msg);

        $has_relations = '';
        if ($this->config->setIfIsSet($related_views,'related_views')) {
            $related_views = explode(',',$related_views);
            $related = array();
            foreach ($related_views as $related_view) {
                if (!$this->config->is_parent("../$related_view")) {
                    continue;
                }
                $view_name = $related_view; 
                $this->config->setIfIsSet($view_name,"../$related_view/display_name");
                $related[$related_view]= $view_name;
            }           
            if (count($related) >0 && ( ($node = $this->template->getElementById('related_views_list', $contentNode)) instanceof DOMNode )) {
                $has_relations = 1;
                foreach ($related as $related_view=>$view_name) {
                    $liNode = $this->template->createElement('li');
                    $linkNode =$this->template->createElement('a',array('href'=>"CustomReports/show/$related_view"), $view_name);
                    $liNode->appendChild($linkNode);
                    $node->appendChild($liNode);
                }
            }
        }
        $this->template->setDisplayDataImmediate('related_views',$has_relations,$contentNode);

        return true;
    }


    private $displayedFields;




    /**
     *Gets the data about the fields we are going to display.
     */
    protected function getDisplayFieldsData() {
        if (is_array($this->displayedFields)) {
            return $this->displayedFields;
        }
        if (array_key_exists('displayFields',$this->defaultOptions)) {
            $displayFields = $this->defaultOptions['displayFields'];
        } else {
            $displayFields = array();
        }
        if (!is_array($displayFields) || count($displayFields) == 0) { //display field information has not been explicitly set
            $fieldData = $this->getReportViewDisplayedFields(false); //don't get disabled fields.
            $has_total = $this->defaultOptions['total'];
            if ($has_total) {
                $fieldData['total'] = array('header'=>'Total','link'=>false, 'link_append'=>false);            
            }
        } else { //we explicility set some fields
            $t_fieldData = $this->getReportViewDisplayedFields(true,true); // get disabled fields.
            $fieldData = array();
            $has_total = false;
            foreach ($displayFields as $displayFieldData) {
                $reportformfield = $displayFieldData['formfield'];
                if (!$reportformfield) {
                    continue;
                }
                if ($reportformfield == 'total') {
                    $fieldData['total'] = array('header'=>'Total','link'=>false, 'link_append'=>false);            
                    continue;
                }
                if (array_key_exists( 'aggregate', $displayFieldData ) && $displayFieldData['aggregate']) {
                    $reportformfield .= '+' . $displayFieldData['aggregate'];
                }
                $t_field = $reportformfield;
                while( $t_field && 
                        !array_key_exists( $t_field, $t_fieldData ) ) {
                    $t_field = substr( $t_field, 0, strrpos( $t_field, '+' ) );
                }
                $fieldData[$reportformfield] = $t_fieldData[$t_field];
            }
        }
        $this->displayFields = $fieldData;
        return  $this->displayFields;
    }


    /**
     * Returns an array in (default display order) with keys reportformfields and values an array with indices 'header', 'link'
     * The value of 'header' is the display header for this field
     * The value of 'link' is false if there is no linked page associtated to this field or a string if there is.
     * @param boolean $getDisabled Defaults to false;
     * @param Mixed $all_aggerates Defaults to false.  If true we include all information about aggregate and non-agregate versions of the fields.
     * otherwise it can be an array of aggregate information that we want
     * @returns array;
     */
    protected function getReportViewDisplayedFields($getDisabled = false, $all_aggregates = false) {
        $displayedFields = array();
        $fieldBaseConfig = $this->config->fields;
        $formfields = $fieldBaseConfig->getKeys();

        foreach ($formfields as $formfield) { //a $formfield of the form '+XXXX' is a function field.  of the form 'XXX+YYY' is a reportform field
            if (!$getDisabled) { //we dont get disabked
                $enabled = false;
                $this->config->setIfIsSet($enabled,"fields/$formfield/enabled");
                if (!$enabled) {
                    continue;
                }
            }
            list($form,$field_agg) = array_pad(explode('+',$formfield),2,'');            
            list($field,$aggregate) =  array_pad(explode('+',$field_agg),2,'');           
            $link = false;
            $link_append = '';
            if (!$form)  { //this is  function field and the name of the field is stored in 
                $header =  I2CE_MagicDataExport_Template::humanText($field);
                $this->reportConfig->setIfIsSet($header,"reporting_functions/$field_agg/header");
            } else { //this is a reportform+field
                $header = I2CE_MagicDataExport_Template::humanText($form) . ' ' . I2CE_MagicDataExport_Template::humanText($field); 
                $this->reportConfig->setIfIsSet($header,"reporting_forms/$form/fields/$field_agg/header");
                $this->reportConfig->setIfIsSet($link,"reporting_forms/$form/fields/$field_agg/link");            
                $this->reportConfig->setIfIsSet($link_append,"reporting_forms/$form/fields/$field_agg/link_append");            
            }
            if ($all_aggregates === true) {
                $aggs = array('sum','maximum','minimum','average','count','');
                $displayedField = $form . '+' . $field;
            } elseif ( is_array( $all_aggregates ) ) {
                $aggs = $all_aggregates;
                $displayedField = $form . '+' . $field;
            } else { // get default aggregate stuff
                if ( !$aggregate ) {
                    $aggreate = '';
                    $this->config->setIfIsSet( $aggregate, "fields/$formfield/aggregate" );
                }
                if ( $aggregate == 'none' ) {
                    $aggregate = '';
                }
                $aggs = array($aggregate);
                $displayedField = $formfield;
            }
            foreach ($aggs as $agg) {
                $t_link = false;
                $t_link_append = false;
                $t_header = $header;
                $t_field = $displayedField;
                switch($agg) {
                case 'sum':
                    $t_header .= ' (Sum)';
                    break;
                case 'maximum':
                    $t_header .= ' (Maximum)';
                    break;
                case 'minimum':
                    $t_header .= ' (Minimum)';
                    break;
                case 'average':
                    $t_header .= ' (Average)';
                    break;
                case 'count':
                    $t_header .= ' (Total)';
                    break;
                case '' :
                case 'none' :
                    $agg = '';
                    $t_link = $link;
                    $t_link_append = $link_append;
                    break;
                default: //badness
                    continue 2;
                    break;
                }
                //if ( $agg && !$aggregate ) {
                //  $t_field .= '+' . $agg;
                //}
                if ($agg) {
                    $t_field .= '+' . $agg;
                }
                $displayedFields[$t_field] = array('header'=>$t_header,'link'=>$t_link, 'link_append'=>$t_link_append);            
            }
        }
        $displayOrder = '';
        if (array_key_exists('display_order',$this->defaultOptions) && $this->defaultOptions['display_order']) {
            $displayOrder = $this->defaultOptions['display_order'];
        } else {
            $this->config->setIfIsSet($displayOrder,'display_order');
        }
        $displayOrder = explode(',',$displayOrder);
        $orderedFields = array();
        foreach ($displayOrder as $field) {
            if (!$field) {
                continue;
            }
            foreach ($displayedFields as $fieldName=>$details) {
                if (strpos($fieldName,$field) !== 0) {                    
                    continue;
                }
                if ( (strlen($fieldName) == strlen($field)) || $fieldName[strlen($field)] == '+') {
                    $orderedFields[$fieldName] = $details;
                    unset($displayedFields[$fieldName]);
                }
            }
        }
        while (count($displayedFields) > 0) {
            reset($displayedFields);
            $key =key($displayedFields);
            $orderedFields[$key] = array_shift($displayedFields);
        }
        return $orderedFields;
    }

  }
# Local Variables:
# mode: php
# c-default-style: "bsd"
# indent-tabs-mode: nil
# c-basic-offset: 4
# End:
