<?php
/**
 * @copyright © 2008, 2009 Intrahealth International, Inc.
 * This File is part of I2CE
 *
 * I2CE is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
*  I2CE_CustomReport_Display_Export
* @package I2CE
* @subpackage Core
* @author Carl Leitner <litlfred@ibiblio.org>
* @version 2.1
* @access public
*/


class I2CE_CustomReport_Display_Export extends I2CE_CustomReport_Display{

    protected static $col_search = 
        array('tab'=>array("\r","\n","\t"),
              'csv'=>'"',
              'html'=>''
            );
    protected static $col_replace = 
        array('tab'=>array("","","        "),
              'csv'=>'""',
              'html'=>''
            );
    protected static $col_seperate = 
        array('tab'=>"\t",
              'csv'=>',',
              'html'=>''
            );
    protected static $col_begin = 
        array('tab'=>'"',
              'csv'=>'"',
              'html'=>'<td>'
            );
    protected static $col_end = 
        array('tab'=>'"',
              'csv'=>'"',
              'html'=>'</td>'
            );
    protected static $row_begin = 
        array('tab'=>'',
              'csv'=>'',
              'html'=>'<tr>'
            );
    protected static $row_end = 
        array('tab'=>"\n",
              'csv'=>"\n",
              'html'=>"</tr>\n"
            );

    /**
     * Process a result row.
     * @param array $row
     * @param int $row_num The current row number when processing results.  If there was a result limit, it starts the count from the beginning of the
     * result offset.  Othwerwise, it starts counting form zero.
     * @param DOMNode $contentNode. Default to null. A node to append the result onto
     */
    protected function processResultRow($row,$row_num,$contentNode=null) {
        $mapped_row = $this->mapResults($row);
        if ($this->style == 'xml') {
            $this->processResultRowXML($mapped_row,$row_num);
        } else {
            array_unshift( $mapped_row, $row_num );
            $this->processResultRowArray( $mapped_row );
        }
        if ( $row_num % 500 == 0 ) ob_flush();
        return true;
    }

    protected function processResultRowArray($row) {
        $vals = array();

        foreach ($this->cols as $key) {
            if ($key == '#') {
                continue;
            }
            if (array_key_exists($key,$row)) {
                $vals[] = self::$col_begin[$this->style] . str_replace(self::$col_search[$this->style],self::$col_replace[$this->style],$row[$key]) . self::$col_end[$this->style]  ;
            } else {
                $vals[] = '';
            }
        }
        echo self::$row_begin[$this->style] . implode(self::$col_seperate[$this->style],$vals) . self::$row_end[$this->style];
    }

    protected function canView() {
        return true;
    }

    protected function processResultRowXML($row, $row_num) {
        $out = " <dataRow id='$row_num'>\n";
        $vals = array();
        foreach ($this->cols as $key) {            
            if (array_key_exists($key,$row)) {
                $val = $row[$key];
            } else {
                $val = '';
            }
            $out .= "  <dataElement name='$key'>" . htmlentities($val) . "</dataElement>\n";
        }
        echo $out . " </dataRow>\n";
    }


    protected function getXMLMetaData($headers) {
        $user = new I2CE_User();
        $sender_role = $user->getRole();
        $sender_username = $user->username();
        $sender_name =   $user->firstname . ' ' . $user->lastname;;
        $sender_email = $user->email;
        $data_elements = '';
        $report = addslashes($this->view );
        $name = $report;
        $this->config->setIfIsSet($name,'display_name');
        $desc = $name . ' report';
        $this->config->setIfIsSet($desc,'description');
        $when = date("c");
        array_shift($headers); //get rid of the # header                                     

        

        foreach ($headers as $key=>$header) {  
            $data_elements .= "   <elemDesc id='$key'>
    <name>$header</name>
   </elemDesc>\n";
        } 
        return "<ihrisReport id='$report'> 
 <reportDetails>
  <name>$name</name>
  <description>$desc</description>
  <whenGenerated>$when</whenGenerated>
  <sender>
   <user>$sender_username</user>
   <name>$sender_name</name>
   <role>$sender_role</role>
   <email>$sender_email</email>   
  </sender>
  <dataElements>
$data_elements  </dataElements>
 </reportDetails>
 <reportData>
";
        return $out;
    }

   
    protected $cols;

    protected $style;

    /**
     * Display the report
     * @param DOMNode $contentNode The DOM node we wish to display into
     * @param boolean $processResults Defaults to true meaning we run through the results
     * @param mixed $controls.  If null (default), we display all the report controsl.  If string or an array of string, we only display the indicated controls
     * @returns boolean. true on sucess
     */
    public function display($contentNode,$processResults=true, $controls = null) {        
        I2CE::longExecution( array( "max_execution_time" => 1800 ) );
        $this->saveDefaultView();
        $this->style = strtolower($this->defaultOptions['export_style']);
        if ( !$this->style || !(array_key_exists($this->style, self::$col_search) || $this->style == 'xml')) {
            $this->style = 'csv';
        }
        $formfields = $this->getDisplayFieldsData();
        $headers = array('#');
        foreach ($formfields as $formfield=>$data) {
            $headers[$formfield] = $data['header'];
        }
        $this->cols = array_keys($headers);
        $errors = '';
        while  (ob_get_level() > 0) {
            $errors .= ob_get_contents();
            ob_end_clean();
        }
        if ($errors) {
            I2CE::raiseError("Got errors:\n$errors");
        }
        $filename = addslashes(str_replace(array(' ',"\n","\t") , array('_',' ','_'),$this->config->display_name)) . '_' . date("d_m_Y");
        switch ($this->style) {
        case 'xml':
            $post = " </reportData>\n</ihrisReport>\n";
            header("Content-disposition: attachment; filename=\"$filename.xml\"");
            header("Content-type: text/xml; charset=UTF-8");  
            array_shift($this->cols);
            echo $this->getXMLMetaData($headers);
            break;
        case 'tab':
            $post = '';       
            if (preg_match('/\s+MSIE\s+\d\.\d;/',$_SERVER['HTTP_USER_AGENT'])) {
                $windows = true;
                header("Content-type: application/vnd.ms-excel");
            } else {
                $windows = false;
                header("Content-type: text/csv; charset=UTF-8"); 
            }
            header("Content-disposition: attachment; filename=\"$filename.csv\"");
            $this->processResultRowArray($headers);
            break;
        case 'csv':
            $post = '';
            if (preg_match('/\s+MSIE\s+\d\.\d;/',$_SERVER['HTTP_USER_AGENT'])) {
                header("Content-type: application/vnd.ms-excel");
            } else {
                header("Content-type: text/csv; charset=UTF-8");  
            } 
            header("Content-disposition: attachment; filename=\"$filename.csv\"");
            $this->processResultRowArray($headers);
            break;
        default: //html snippet
            header("Content-disposition: attachment; filename=\"$filename.html\"");
            foreach( $this->template->getHeaders() as $header) {
                header($header);
            }
            flush();
            $name = addslashes(str_replace(array(' ',"\n","\t") , array('_',' ','_'),$this->config->display_name));
            echo "<table id='" .  addslashes($this->view )." name='$name'>\n";
            $post = '</table>';  
            $this->processResultRowArray($headers);
            break; 
        }
        // Flush the headers so the download box appears fast
        flush();
        $data = $this->getResults();
        ob_start();        
        if (!$this->processResults($data,$resultsNode=null)) {
            I2CE::raiseError("Could not get results");
        }         
        echo $post;
        flush();
        exit; // we want to make sure there is no further output or that the $this->page->display() method is not called
    }






}
# Local Variables:
# mode: php
# c-default-style: "bsd"
# indent-tabs-mode: nil
# c-basic-offset: 4
# End:
